%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Db.h>
#include <DbDefs.h>
#include <DbObject.h>
#include <DbTypeParse.h>

typedef struct
{
  char* buffer;
  int current_position;
  int line_number;
  char* old_line_start;
  char* line_start;
} LexContext;

int DbTypeLexLineNumber ();
char* DbTypeLexLineStart ();
int DbTypeLexPos ();
LexContext* DbTypeLexStartup (LexContext* context, char* text);

#ifdef YYLMAX
#undef YYLMAX
#endif

#define YYLMAX 2000

int DbTypeLexInput ();
void DbTypeLexUnput (int c);

#ifndef YY_NO_INPUT

#undef input
#define input DbTypeLexInput
#undef unput
#define unput DbTypeLexUnput

#else

#define internal_input DbTypeLexInput

#undef YY_INPUT
#define YY_INPUT(buffer,result,max_size) \
{ \
  int c = '*', n; \
  for ( n = 0; n < max_size && \
               ((c = internal_input()) != 0) && \
               (c != '\n'); ++n ) \
    { \
      buffer[n] = (char) c; \
    } \
  if ( c == '\n' ) buffer[n++] = (char) c; \
  result = n; \
}

#endif


static LexContext* Context = NULL;

static char* StackPut (char* text);

int DbTypeYYlex ();

%}

%start DB

space     [ ;\t\r\f]
alphanum  [a-zA-Z0-9][a-zA-Z0-9_\-]*
term      [^a-zA-Z0-9]

%%

<DB>"#"[^\n]*\n {
  Context->old_line_start = Context->line_start;
  Context->line_start = &Context->buffer[Context->current_position];
  Context->line_number++;
}

<DB>\n {
  Context->old_line_start = Context->line_start;
  Context->line_start = &Context->buffer[Context->current_position];
  Context->line_number++;
}

<DB>{space}+ {
}

<DB>"{" {
  return (OPEN_BRACE);
}

<DB>"}" {
  return (CLOSE_BRACE);
}

<DB>"[" {
  return (OPEN_SQUARE);
}

<DB>"]" {
  return (CLOSE_SQUARE);
}

<DB>"<" {
  return (LOWER);
}

<DB>">" {
  return (GREATER);
}

<DB>"," {
  return (COMMA);
}

<DB>"structure" {
  return (STRUCTURE);
}

<DB>"short" {
  return (SHORT);
}

<DB>"int" {
  return (INT);
}

<DB>"hexa" {
  return (HEXA);
}

<DB>"float" {
  return (FLOAT);
}

<DB>"double" {
  return (DOUBLE);
}

<DB>"string" {
  return (STRING);
}

<DB>"reference" {
  return (REFERENCE);
}

<DB>"char" {
return (CHARACTER);
}

<DB>([+-])?[0-9][0-9]*/{term} {
  int n;

  sscanf ((char*) yytext, "%d", &n);

  yylval.number = n;
  return (NUMBER);
}

<DB>\'[^']*\'/{term} {
  char c;

  sscanf ((char*) &yytext[1], "%c", &c);

  yylval.number = c;

  return (NUMBER);
}

<DB>{alphanum} {
  char* string;

  string = (char*) &yytext[0];

  yylval.string = StackPut (string);

  return (NAME);
}

%%

/*---------------------------------------------------------------*/
int DbTypeLexLineNumber ()
/*---------------------------------------------------------------*/
{
  return (Context->line_number);
}


/*---------------------------------------------------------------*/
char* DbTypeLexLineStart ()
/*---------------------------------------------------------------*/
{
  return (Context->line_start);
}

/*---------------------------------------------------------------*/
int DbTypeLexPos ()
/*---------------------------------------------------------------*/
{
  int pos;

  pos = Context->line_start - &Context->buffer[Context->current_position];

  return (pos);
}

/*---------------------------------------------------------------*/
int DbTypeLexInput ()
/*---------------------------------------------------------------*/
{
  char c;

  if (!Context->buffer) return (0);

  if (Context->current_position < strlen(Context->buffer))
    {
      c = Context->buffer[Context->current_position];
      Context->current_position++;
    }
  else c = 0;

  return (c);
}

/*---------------------------------------------------------------*/
void DbTypeLexUnput (int c)
/*---------------------------------------------------------------*/
{
  if (!Context->buffer) return;

  if (Context->current_position > 0)
  {
    Context->current_position--;
    Context->buffer[Context->current_position] = c;
  }
}

/*---------------------------------------------------------------*/
void DbTypeLexRestoreContext (LexContext* context)
/*---------------------------------------------------------------*/
{
  if (Context->buffer != NULL)
    {
      free (Context->buffer);
      Context->buffer = NULL;
    }

  Context = context;
}

/*---------------------------------------------------------------*/
LexContext* DbTypeLexStartup (LexContext* context, char* text)
/*---------------------------------------------------------------*/
{
  LexContext* previous = Context;

  Context = context;

  yyin = stdin;
  yyout = stdout;

  Context->buffer           = NULL;
  Context->current_position = 0;
  Context->line_number      = 0;
  Context->old_line_start   = NULL;
  Context->line_start       = NULL;

  Context->buffer = (char*) malloc (strlen (text) + 2);
  strcpy (Context->buffer, text);
  strcat (Context->buffer, " ");

  Context->current_position = 0;
  Context->line_number = 1;
  Context->old_line_start = Context->buffer;
  Context->line_start = Context->buffer;

  BEGIN DB;

  return (previous);
}

/*---------------------------------------------------------------*/
static char* StackPut (char* text)
/*---------------------------------------------------------------*/
{
#define MAXDEPTH 150

  static struct
  {
    char text[YYLMAX];
  } stack[MAXDEPTH];

  static int index = 0;

  char* str;

  str = (char*) stack[index].text;
  strcpy (str, text);

  index++;

  if (index >= MAXDEPTH) 
  index = 0;

  return (str);
}
